enum AlgState {BACKTRACKING,BCP,BCP_CNFL,BPE,DECISION_MAKING,FINISH,RESOURCE_EXCEEDED}; // see n1otes
enum BranchType {LEFT,RIGHT};
enum BCPreturn {CONFLICT,NON_CONFLICT}; 
enum BCPfirstNode {DEC_LEFT,DEC_RIGHT,IMPLIED,SPEC_IMPLIED};
enum XvarDecMaking {XVAR_RAND,XVAR_TRIV,XVAR_MAX_ACT_SGNL_CUBES,XVAR_MAX_ACT_CUBES,XVAR_LIT_ACTIVITY};
enum YvarDecMaking {YVAR_RAND,YVAR_TRIV,YVAR_MAX_ACT_CUBES,YVAR_LIT_ACTIVITY3,YVAR_LIT_ACTIVITY5};
enum CubeType {INITIAL,Y_CUBE,X_CUBE,NONE};
enum DseqType {SAT,UNSAT,UNRES,LATE_JOIN,UNS_JOIN,LATE_UNS_JOIN,JOIN}; // Types of Dsequents the first three are atomic, the last is not
           // SAT - cube is satisfied, UNSAT - cube is subsumed
           // UNRES -cube cannot be resolved, JOIN - D-sequent is obtained by the join operation
           // LATE_JOIN - D-sequent of a late cube obtained by a "special" join operation
           // UNS_JOIN  - D-sequent obtained by join operation of a D-sequent and a cube
           // LATE_UNS_JOIN - UNS_JOIN for a late cube

const int MAX_NAME = 100;


typedef vector<int> CUBE;
typedef vector<CUBE> DNF;
typedef CUBE CLAUSE;
typedef DNF CNF;
typedef set <int> SCUBE;
typedef unsigned int uint;
typedef map<int,int>  CubeMap;
typedef map<CUBE,bool> SeqMap;
typedef vector <char> CCUBE;
typedef deque <int> DCUBE;
typedef Minisat::vec<Minisat::Lit> TrivMclause;
typedef Minisat::Lit Mlit;
typedef TrivMclause MvecLits;
typedef Minisat::Clause ComplMclause;
typedef Minisat::lbool Mbool;
typedef Minisat::vec<Mbool> MboolVec;
typedef Minisat::Solver Msolver;
const Mbool Mtrue = Minisat::l_True;
const Mbool Mfalse = Minisat::l_False;
const Mbool Mundef = Minisat::l_Undef;
		   
class hsh_tbl 
{public:  
 int marker;
 CUBE table;
 bool in_use; // is a semafor to prevent using a hash table by two different agents
  //
  // methods
  //
 void  change_marker(void);
 void ht_init(int nelems);
 void clean();
 void add_elem(void);
 int size();
 void started_using(void);
 void done_using(void);
}; 

struct VarState
{  
  unsigned val : 2; // 0 - false, 1 - true, 2 - unassigned
  unsigned xvar : 1; // 1 means that this variable is in the set X i.e. quantified
                     // 0 means that the variable is in the set Y (i.e. unquantifed)

  unsigned in_array : 1; //  this field is set to 1 if the X-variable is in the array 'Sgnl_xvars'
                         //  regardless of whether it is assigned 
};


struct Node
{int var; // the variable selected for branching at this node

 char lval; // the first value assigned to 'var' (i.e. the value of the left branch)
 BranchType branch; // takes values LEFT and RIGHT 
 int left_ecube; // if left_ecube >= 0, then it specifies  the index of the empty cube
                             // cube derived in the left child node. Otherwise, left_ecube=-1
 int right_ecube; // if right_ecube >= 0, then it specifies  the index of the empty cube
                             // cube derived in the right child node. Otherwise, right_ecube=-1 
};

struct CubeInfo
{  
  char  state;    // state of the cube
                  // ACTIVE, cube is active
                  // SATISFIED, cube is inactive becaise it is satisfied
                  // REDUNDANT, cube is inactive because it is prove redundant in the current subspace
  CubeType type;  // type of the cube
                  //  Y_CUBE, cube  contains only Y-variables
                  //  X_CUBE cube contains at least one X-variable
                  //  NONE - cube is empty
                  // INITIAL - not type has been assigned yet after initialization

  unsigned signal : 1; // 1  means that this is an X-cube that is currently have the signal status

  unsigned glb_sgnl:1; // 1 means that this is an X-cube that is currently in Prim_sgnl_cubes

  int depth; // If (signal == 1), then 'depth' specifies the depth up to which the cube remains
              // its signal status
};
