#include <iostream>
#include <list>
#include <vector>
#include <queue>
#include <set>
#include <map>
#include <stdio.h>
#include <stdlib.h>
#include <strings.h>
#include <algorithm>
#include <ctime>
#include <climits>
#include <cassert>
#include <string.h>

#include "Solver.h"
#include "SimpSolver.h"
#include "System.h"
using namespace std;
#include "e0g-pqe+.hh"
#include "einline.hh"


/*====================================

  B P E _ S T A T E

  =====================================*/
AlgState ExQuant::bpe_state(CUBE &Recov_act_cubes)
{

  merging_nodes++;  
  Node &n = Tree_path[curr_depth];

 
  bool cond_holds = check_spec_cond(n);
  if (cond_holds)   {
    if (verbose) printf("condition!!\n");
    finish_proc_sat_cubes();
    Assgn_queue.clear();
    int assgn = (n.lval == 0)?-n.var:n.var;
    push_assgn(Assgn_queue,n.right_ecube,assgn);
    bcp_first_node = SPEC_IMPLIED;
    delete_node(curr_depth);
    return(BCP);
  }


  
 
  join_dseqs(n); 
  finish_proc_sat_cubes();

 
  CUBE Late_cubes;
  new_late_cubes(n,Recov_act_cubes,Late_cubes);
  if (Late_cubes.size() > 0)   {
    assert(n.right_ecube >= 0);
    make_cubes_red(n,Late_cubes);
  }
    
  if (xvar_check(n)) seq_gen_ecube(n);

  delete_node(curr_depth);  
 
  // switch on the yvar mode if appropriate
  update_var_mode();
  if (yvar_mode) return(FINISH);
  return(BACKTRACKING); 
  
} /* end of function bpe_state */


/*===========================================

  M A K E _ C U B E S _ R E D

  This function is called to update D-sequents
  of 'Late_cubes' (i.e. derived cubes 
  that have no D-sequent in the left branch)

  ASSUMPTIONS:
  1) n.left_child_cube < 0
  2) n.right_child_cube >= 0
  3) Every cube F[Late_cubes[i]] is in 
  state ACTIVE

  ============================================*/
void ExQuant::make_cubes_red(Node &n,CUBE &Late_cubes)
{

  CUBE &C = F[n.right_ecube];
  CUBE NewSeq;
  int right_lit;

  if (n.lval == 0) right_lit = n.var;
  else right_lit = -n.var;
    
  bool ok = quick_res(NewSeq,C,right_lit);
  assert(ok);      

  for (uint i=0; i < Late_cubes.size(); i++)   {
    int cube_ind = Late_cubes[i];
    assert(Cube_info[cube_ind].state == ACTIVE);

    int L = max_cube_recovery_depth(NewSeq);
    set_new_act_dseq(cube_ind,NewSeq,false,LATE_UNS_JOIN);	 
    check_red_cube_in(cube_ind,0,L,true);
    num_fc_dseqs++;
  }

} /* end of function make_cubes_red */

/*==========================================


  N E W _ L A T E _ C U B E S 

  This function returns 'true'  if 
  'Recov_act_cubes' contains index of 
  a late cube. 


  ==========================================*/
void ExQuant::new_late_cubes(Node &n,CUBE &Recov_act_cubes,CUBE &Late_cubes)
{

 
  int lbranch_size = Formula_size[n.var-1];
  for (uint i=0; i < Recov_act_cubes.size(); i++)   {
    int cube_ind = Recov_act_cubes[i];
     
    if (Cube_info[cube_ind].signal == 0) continue; // not a signal cube
  
    if (lbranch_size >= cube_ind+1)	{
      if (Cube_info[cube_ind].state == ACTIVE)   { 
        p();
        printf("a non-late cube  %d is active\n",cube_ind);
        exit(ERROR);
      }
      continue;  // this cube is not late
    }
      
    if (Cube_info[cube_ind].state != ACTIVE) continue; // this cube is not active

    Late_cubes.push_back(cube_ind);
  }


} /* end of function new_late_cubes */


/*===========================================

  F I N I S H _ P R O C _ S A T _ C U B E S

  This function takes care of Y-cubes
  that become active after unassigning
  the branching variable. 

  ============================================*/
void ExQuant::finish_proc_sat_cubes()
{

  CUBE &Cubes = New_sat_cubes[curr_depth];
  for (uint i=0; i < Cubes.size(); i++)  {
    int cube_ind = Cubes[i];
    if (Cube_info[cube_ind].type != Y_CUBE) continue;
    Cube_info[cube_ind].state = ACTIVE;
    if (Cube_info[cube_ind].signal)  num_act_sgnl_cubes++;
  }
   
  Cubes.clear();

} /* end of function finish_proc_sat_cubes */



/*====================================

  C O P Y _ L I T S

  ====================================*/
void ExQuant::copy_lits(CUBE &toS,CUBE &fromS)
{
 
  for (uint i=0; i < fromS.size(); i++)  {
    int var_ind = abs(fromS[i])-1;
    if (present_in_table(var_ind,fromS[i]))  continue;

    insert_into_table(var_ind,fromS[i]);
    toS.push_back(fromS[i]);
  }


} /* end of function copy_lits */



/*============================================

  A D D _ N E W _ L I T E R A L S

  This function adds to E the literals
  of C that
  a) literals of  X-variables
  b) literals of variables that are not
  already marked in the hash table.

  =============================================*/
inline void ExQuant::add_new_literals(CUBE &E,CUBE &C)
{
  CUBE &table = hash_table_lits.table;
  int marker = hash_table_lits.marker;

  for (uint i=0; i < C.size(); i++) {
    int var_ind = abs(C[i])-1;
    if (Var_state[var_ind].xvar == 0) continue;
    if (table[var_ind] == marker) continue;
    E.push_back(C[i]);
    table[var_ind] = marker;
  }


} /* end of function add_new_literals */
