#include <iostream>
#include <list>
#include <vector>
#include <set>
#include <map>
#include <string>
#include <queue>
#include <algorithm>
//
#include <stdio.h>
#include <assert.h>
#include <string.h>

#include "Solver.h"
#include "SimpSolver.h"
using namespace std;

#include "e0g-pqe+.hh"
#include "einline.hh"

//
// BLOCKING_CLAUSE
//
bool ExQuant::blocking_clause(CLAUSE &B,int cube_ind)
{
  if (verbose) printf("blocking clause\n");   
  fill_yas(Slv);
  init_sgnl_cubes(cube_ind);
  int ans = ds_pqe();
  if (ans == 1) 
    return(false); // unfinished
  form_blocking_clause(B,cube_ind);
  if (verbose) {
    printf("B-> "); print_cube1(B);
    printf("cube_ind = %d\n",cube_ind);
    print_cube_state(cube_ind);
    printf("Red_cube_depth[%d] = %d\n",cube_ind,Red_cube_depth[cube_ind]);
  }
  
  loc_init(B,cube_ind);
  return(true); 
} /* blocking_clause */

//
//  LOCK_INIT
//
void ExQuant::loc_init(CUBE &B,int cube_ind)
{


  Yas.clear();
  alg_state = DECISION_MAKING;
  bcp_first_node = DEC_LEFT;

  delete_tree();

  for (uint i=0; i < Yvars.size(); i++)  {
    int var_ind = Yvars[i]-1;   
    Dseqs_to_check0[var_ind].clear();
    Dseqs_to_check1[var_ind].clear();
  }

  Lit_activity0.assign(num_vars,0);
  Lit_activity1.assign(num_vars,0);

  Var_sel_depth.assign(num_vars,-1);


  for (uint i=0; i < F.size(); i++) {
    CubeInfo &El = Cube_info[i];
    if (El.state == REDUNDANT) 
      if (Red_cube_depth[i] < 0) 
        continue;     
    El.state = ACTIVE;
    Red_cube_depth[i] = num_vars+1;
  }

 

  num_ycubes=0;
  for (uint i=0; i < F.size(); i++) 
    if (Cube_info[i].type == Y_CUBE) 
      num_ycubes++;
   
  dirt_ycubes = num_ycubes;
  init_ycubes = num_ycubes;

  for (uint i=0; i < F.size(); i++)    {
    Left_dseqs[i].clear();
    Sat_vars[i].clear();
    Act_dseqs[i].clear();
  }

  Cause_lit.assign(F.size(),0);

 
  Act_sgnl_cubes_cnt0.assign(num_vars,-1);
  Act_sgnl_cubes_cnt1.assign(num_vars,-1);
  Sgnl_xvars.clear();
  Sgnl_yvars.clear();
  for (int i=0; i < num_vars; i++) 
    Var_state[i].in_array = 0;
 
  set_init_counts();
  num_assgn_xvars = 0;
  num_assgn_yvars = 0;


 
  num_sgnl_xvars = Sgnl_xvars.size();
  num_sgnl_yvars = Sgnl_yvars.size();

  form_unit_ycube_queue();


  init_sgnl_cubes(cube_ind);
  if (verbose) print_sgnl_cubes();

  for (uint i=0; i < F.size(); i++) {
    Ping[i].assign(F[i].size(),-1);
    Ped[i].clear();
  }

  bool solved = assgn_init_prot_cubes();
  if (solved && verbose) {
    printf("**num_act_sgnl_cubes = %d\n",num_act_sgnl_cubes);
    print_sgnl_cubes();
  }
  if (verbose) 
    printf("solved = %d\n",solved);
  
  // rem_red_prim_cubes();
  if (solved) {
    printf("solved = %d\n",solved);
    assert(Cube_info[cube_ind].state == REDUNDANT);
    B = Act_dseqs[cube_ind];
  }

  Formula_size.assign(num_vars,-1);
} /* loc_init */



//
//   FORM_BLOCKING_CLAUSE
//
void ExQuant::form_blocking_clause(CLAUSE &B,int cube_ind)
{

  if (verbose) {
    printf("form_blocking_clause\n");
    cout << "Prim_sgnl_cubes-> " << Prim_sgnl_cubes << endl;
  }

 
  CubeInfo &El = Cube_info[cube_ind];
  if (El.state == ACTIVE) {
    p();  
    exit(ERROR);
  }
  if (El.state == REDUNDANT) 
    B = Act_dseqs[cube_ind];
  else {
    assert(El.state == SATISFIED);   
    add_cause_lit(B,cube_ind);
  }
  

} /* form_blocking_clause */


//
//   FILL_YAS
//
void ExQuant::fill_yas(Msolver *Slv)
{

  MboolVec &S = Slv->model;
  
  for (uint i=0; i < Yvars.size(); i++) {
    int var_ind = Yvars[i]-1;
    if (S[var_ind] == Mtrue) Yas.push_back(var_ind+1);
    else Yas.push_back(-(var_ind+1));
  }
} /* fill_yas */


//
//   DELETE_TREE
//
void ExQuant::delete_tree()
{
  assert(curr_depth == (int) Tree_path.size()-1);
  while (Tree_path.size() > 0)  {
    Node &n = Tree_path.back();
    VarState &Var = Var_state[n.var-1];
    if (Var.xvar) {
      p();
      print_var_state(n.var-1);
      printf("Var.xvar = %d\n",Var.xvar);
      exit(ERROR);
    }
    Var.val = 2;
    New_sat_cubes[curr_depth].clear();
    Left_red_cubes[curr_depth].clear();
    New_sgnl_cubes[curr_depth].clear();
    Tree_path.pop_back();
    curr_depth--;
  }

  curr_depth = -1;

} /* delete_tree */
