#include <iostream>
#include <list>
#include <vector>
#include <queue>
#include <set>
#include <map>
#include <stdio.h>
#include <stdlib.h>
#include <strings.h>
#include <algorithm>
#include <ctime>
#include <climits>
#include <cassert>
#include <string.h>

#include "Solver.h"
#include "SimpSolver.h"
#include "System.h"
using namespace std;
#include "e0g-pqe+.hh" 
#include "einline.hh"


/*=======================================

  P R I N T _ B C P _ U N I T _ C U B E

  ======================================*/
void ExQuant::print_bcp_unit_cube(int unit_cube)
{

  if (unit_cube >= 0)   {
    printf("derived from cube ");
    if (Cube_info[unit_cube].signal) printf("*");
    printf("F[%d]: ",unit_cube);
    print_cube1(F[unit_cube]);
  }

} /* end of function print_bcp_unit_cube */
/*=======================================

  P R I N T _ B C P _ A S S G N

  ======================================*/
void ExQuant::print_bcp_assgn(int assgn)
{

  int var_ind = abs(assgn)-1;
  if (Var_state[var_ind].xvar)
    if (assgn < 0)
      printf("    BCP: making assignment ~x%d at depth %d\n",-assgn,curr_depth);
    else // assgn > 0
      printf("    BCP: making assignment x%d at depth %d\n",assgn,curr_depth);
  else // Y-variable 
    if (assgn < 0)
      printf("    BCP: making assignment ~y%d at depth %d\n",-assgn,curr_depth);
    else // assgn > 0
      printf("    BCP: making assignment y%d at depth %d\n",assgn,curr_depth);

} /* end of function print_bcp_assgn */

/*=================================================

  A D D _ S _ S E Q U E N T S _ D E C _ N O D E


  This function is called in the left branch of
  a decision node 'n'. It forms new active S-sequents
  for the active cubes that got satisfied by
  the current (i.e. left) assignment to 'n.var'


  ASSUMPIONS:
  1) The cubes satisfied by the left assignment
  to 'n.var' are listed in 'New_sat_cubes[curr_depth]'

  =================================================*/
void ExQuant::add_s_sequents_dec_node(int assgn)
{

  CUBE &Cubes = New_sat_cubes[curr_depth];

  CUBE S;
  S.push_back(assgn);

  for (uint i=0; i < Cubes.size(); i++)    {
    int cube_ind = Cubes[i];
    if (Cube_info[cube_ind].signal == 0) continue;
    num_left_s_seqs++;
    num_sc_dseqs++;    
    set_new_act_dseq(cube_ind,S,false,SAT);
  }
 

} /* end of function add_s_sequents_dec_node */


/*======================================================================

  M A K E _ O N E _ A S S I G N M E N T

  The function makes the assignment falsifying a literal 'lit'
  of the cubes listed in 'removal_array'. The cubes from sat_array contain
  literal '~lit' and so they are marked as satisfied.


  All encountered  unit_cubes (and the corresponding assignments)
  are added to assg_queue


  ========================================================================*/
int  ExQuant::make_one_assignment(int assgn,vector <int> &removal_array,vector <int> &sat_array)

{
  int cube_ind;
  int der_assgn;
 
  ///  first check if a cube in the removal array is unit or the universe 
  for (uint i=0; i < removal_array.size();i++)   {
    cube_ind = removal_array[i];
    if (Cube_info[cube_ind].state != ACTIVE)   
      continue; // cube F[cube_ind] is already satisfied or redundant
     
    CUBE &C = F[cube_ind];
    int length = find_length(C,der_assgn);
     
    if (length ==0)  // if an unsatisfied cube is encountered
      return(cube_ind); // return its number

    if (length ==1) {// if a unit cube is encountered
      Assgn_queue.push_back(cube_ind); //add the number of the unit cube to satisfied by this assignment
      Assgn_queue.push_back(der_assgn);  // add the assignment to be made    
    }
  }

  //  mark as satisfied the active cubes set to 0 by assgn

  for (uint i = 0; i < sat_array.size();i++) {
    cube_ind = sat_array[i];
    if (Cube_info[cube_ind].state != ACTIVE)
      continue; // cube cube_ind is either already satisfied or redundant
    check_sat_cube_in(cube_ind,assgn); 
    All_new_sat_cubes.push_back(cube_ind);
  }
  return(-1);
} /* end of function make_one_assignment */




/*==================================================

  R U N _ B C P

  Performs the Boolean constraint propagation
  procedure in the current formula F 
  Returns 

  CONFLIT - if a conflict occured
  NON-CONFLICT if no conflict occured

  ===================================================*/
BCPreturn ExQuant::run_bcp()
{

  bool first_assgn = true;

  while (true)  {
    if (Assgn_queue.size() == 0)
      return(NON_CONFLICT);

    //
    // extract the next assignment to make
    //
  
    int unit_cube_num, assgn;
    pop_assgn(Assgn_queue,unit_cube_num,assgn);
  
    // check if it is a duplicate unit cube  
   
    if (unit_cube_num >= 0) // check that it is a deduced assignment
      {
        if (Cube_info[unit_cube_num].state == SATISFIED) continue; // in case of duplicate unit cubes
        if (Cube_info[unit_cube_num].state == REDUNDANT) continue; // cube has been proved redundant in the process of BCP
      }
 
    //
    // create a new node corresponding to this assignment
    //

    int var_ind = abs(assgn)-1;
    if (yvar_mode) // skip assignments to X-variabls if 'yvar_mode = true'
      if (Var_state[var_ind].xvar) continue;
   
  

    if (first_assgn)    {
      if ((bcp_first_node == DEC_LEFT) || (bcp_first_node == SPEC_IMPLIED))
        create_node(unit_cube_num,assgn);
      else { // DEC_RIGHT or IMPLIED
        make_node_assgn(assgn);
        if (bcp_first_node == IMPLIED)
          if (Cube_info[unit_cube_num].signal) {
            add_bcp_sgnl_cubes(assgn,curr_depth);
          }
      }
      first_assgn = false;
    }
    else // not first assignment
      create_node(unit_cube_num,assgn);
   
  
    if (verbose) {
      print_bcp_assgn(assgn);
      print_bcp_unit_cube(unit_cube_num);      
    }

    
  
    if (Var_state[var_ind].xvar) // num_sgnl_xvars is decreased by 'decrease_sgnl_cube_scores'
      num_assgn_xvars++;
    else
      num_assgn_yvars++;
    
   
    //
    //  update cubes
    //
    int unsat_cube;
  
   
    if (assgn < 0)
      unsat_cube = make_one_assignment(assgn,Flits0[var_ind],Flits1[var_ind]);
    else
      unsat_cube = make_one_assignment(assgn,Flits1[var_ind],Flits0[var_ind]); 
      
    if (unsat_cube < 0)   {
      Node &n = var_node(var_ind);
      if (n.branch == LEFT) add_s_sequents_dec_node(assgn);	 
      continue; // no  cube universe  has been produced  yet
    }
    else {
      Assgn_queue.clear();      
      empty_cube_index = unsat_cube;
      return(CONFLICT);
    }
  }
  
} /* end of procedure run_bcp */


/*===========================

  B C P _ S T A T E

  ============================*/
AlgState ExQuant::bcp_state()
{ 

  if (bcp_first_node != IMPLIED) // When a decision making procedure employs a unit cube,                                  
    All_new_sgnl_cubes.clear();  //it fills up  'All_new_sgnl_cubes'.
  All_new_sat_cubes.clear();
  

  BCPreturn answer = run_bcp();
  if (answer == NON_CONFLICT)    {
    CUBE Red_cubes;
    confirm_sgnl_cubes(All_new_sgnl_cubes,Red_cubes,curr_depth);
    check_for_new_red_cubes(Red_cubes);
    check_for_new_red_cubes(All_new_sat_cubes);
    return(DECISION_MAKING); // no conflict occured
  }
  else  {
    num_confls++;
    return(BCP_CNFL);
  }

} /* end of function bcp_state */
