#include <iostream>
#include <list>
#include <vector>
#include <queue>
#include <set>
#include <map>
#include <stdio.h>
#include <stdlib.h>
#include <algorithm>
#include <ctime>
#include <climits>
#include <cassert>
#include <string.h>

#include "Solver.h"
#include "SimpSolver.h"
#include "System.h"
using namespace std;
#include "e0g-pqe+.hh"
#include "einline.hh"

long gcount = 0;
bool dec_making_flag;
int cnfl_cube_flag;

/*============================================

       D S _ P Q E

  0 - finished
  1 - unfinished
  2 - preprocessing

  ASSUMPTIONS:
  1) F does not have an empty cube
  ==============================================*/
int ExQuant::ds_pqe()
{


  CUBE Recov_act_cubes;
  prev_alg_state = alg_state;
  dec_making_flag = 0;
  if (Yas.size() > 0) yvar_mode = 1;

  if (verbose)
    printf("calling DS-PQE\n");


  while (true)    { 
     
    set_verbose();
    if (verbose)     {	 
      print_alg_state(true);
      if (gcount == start_gcount) my_printf((char *) "*gcount = %m\n",gcount);
      else my_printf((char *) "gcount = %m\n",gcount);
    }
 
    gcount++;
      
 
      
    if (resource_exceeded(gcount)) alg_state = RESOURCE_EXCEEDED;    
   
    AlgState tmp_state;
    switch(alg_state) 	{
    case FINISH:
      return(0);
    case RESOURCE_EXCEEDED:
      return(1);
    case DECISION_MAKING:
      if (yvar_mode)  tmp_state = dec_making_state_yvar();
      else  tmp_state = dec_making_state_xvar();
      break;
    case BCP:
      tmp_state = bcp_state();
      break;
    case BCP_CNFL:
      tmp_state = bcp_cnfl_state(Recov_act_cubes);
      break;
    case BPE:     
      tmp_state = bpe_state(Recov_act_cubes); 
      Recov_act_cubes.clear();
      break;
    case BACKTRACKING:   
      if (xvar_check(Tree_path[curr_depth]))  tmp_state = backtrack_state_xvar();
      else  tmp_state = backtrack_state_yvar();
      break;
    default:
      assert(false);
    }
    prev_alg_state = alg_state;
    alg_state = tmp_state;
  }

} /* end of function ds_pqe */

/*====================================

  R E S O U R C E _ E X C E E D E D

  ===================================*/
inline bool ExQuant::resource_exceeded(long int gcount)
{
 
  if ((max_num_ycubes >= 0) && (dirt_ycubes-init_ycubes > max_num_ycubes)) {
    printf("num_ycubes = %d, init_ycubes = %d\n",num_ycubes,init_ycubes);
    return(true);
  }
  if ((gcount  & 0xff) == 0)      {
    double usrtime,systime;
    get_runtime (usrtime, systime);
    if (usrtime > time_limit) return(true);
  }
  return(false);
  
} /* end of function resource_exceeded */


/*=====================================

  F O R M _ R E S U L T

  ======================================*/
void ExQuant::form_result(DNF &F1)
{ 

  F1.clear();
  if (F.back().size() == 0) {
    CUBE C;
    F1.push_back(C);
    return;
  }
 
  for (uint i=0; i < F.size(); i++)   {
    CUBE &C = F[i];
    if (Xcube(C)) continue;
    if (noise == false) 
      if (Clean[i]) continue;
    F1.push_back(C);
  }

} /* end of function form_result */
