#include "aux_types.hh"


/*============================

  C L A S S   E X Q U A N T

==============================*/
class ExQuant
{
public:
  DNF Ybps; // boundary points in terms of yvar assignments
  long int last_gcount; // the last algorithm's state
  DNF F; // current DNF formula
  DNF Sol; // solution to the PQE problem
  DCUBE Prim_sgnl_cubes; // describes the current primary signal cubes of F
                         // i.e. the cubes to be taken out of F
  CNF F2; //  F minus the initial signal cubes
  CUBE Trgs; // target clauses of F
  DCUBE Yas; // assignments corresponding to the current unremovable
            // boundary point
  SCUBE New_prim_scubes; // new primary signal cubes
 
  
  Msolver *Slv; // Sat-solver used for generation of boundary points
                  // to make decisions when the current target is the primary one

  CCUBE Xvar; // Xvar[i] = 0 (1) if variable i+1 is in Yvars
             // (resp. not in Yvars)

  int rem_bnd_pnts; // count of removable boundary points
  int unrem_bnd_pnts; // count of unremovable boundary points
  int max_new_ycubes; // limit on the number of new y-cubes generated
  int new_ycubes; // number of ycubes generated so far
  long int max_gcount; // terminate after max_gcount is reached
  
  // D-sequent related class members

  
  DNF Act_dseqs;// Act_dseqs[i] is the current active sequent of the cube F[i]
  vector <DNF> Left_dseqs; // Left_dseqs[i] gives the current stack of left (i.e. waiting) sequents for cube F[i]
 
  DNF Sat_vars; // Sat_vars[i][j] specifies the var of j-th sequent for i-th cube that got
               // satisfied (and so was the reason why the sequent Left_dseqs[i][j] became inactive
  DNF Left_red_cubes; // 'Left_red_cubes[depth]' specifies the cubes that became non-redundant due
                      //  to flipping the value of variable 'Tree_path[depth].var'. In particular,
                      // 'Left_red_cubes[depth]' contain the cubes satisfied by the left assignment to 'Tree_path[depth].var'.
   
  

  CUBE Cause_lit;   // If cube F[i] gets satisfied, Cause_lit[i] contains the negation of the literal 
                            // that gets satisfied. If F[i] is redundant, then 
                            // if the D-sequent of F[i] is atomic, Cause_lit[i] gives the literal of F[i] that is unprotected
                            // if the D-sequent of F[i] is obtained by resolution, then Cause_lit[i] = 0

 
                            
  vector <CubeMap> Dseqs_to_check0;  //  Dseqs_to_check0[(index(x)] specifies the set Z of cube indexes  such that for every 
                                     // cube z of Z there is an active sequent (,..,x=0,...,) -> {Cube(z)}

  vector <CubeMap> Dseqs_to_check1;  // The same as 'Dseqs_to_check0' but for positive literals.

 

  // new data members for bookkeeping of redundant cubes

  
  DNF Ping; // Ping[i] consists of 'F[i].size()' elements. Ping[i][j] specifies index of a cube C* of F that protects F[i] from removal.
            // C* and F[i] are resolvable on variable abs(C[j]),
            //  Ping[i][j] is equal to -1 in three cases: a) when this element is initialized, b) when F[i][j] is an Y-variable
            // c) When cube F[i] is not resolvable on variable F[i][j]  with any other cube in F (this may happen for derived cubes)

  DNF Ped; // Ped[i] specifes pairs (cube_ind,lit_ind) describing cubes protected by cube C=F[i]. Namely, F[cube_ind] is a
           // cube protected by C and the former and the latter are resolvable on variable abs(F[cube_ind][lit_ind])
                       
  // other class members


  AlgState alg_state;
  AlgState prev_alg_state;
  int first_yvar_assigned;
  bool yvar_mode; // is set to true if at least one variable of Y is unassigned and non-redundant
 

  vector <Node> Tree_path;// Tree_path[i] is the node corresponding to selection depth i
  // Selection depth 0 corresponds to the root node
  // Selction and assignment depths of a variable are equival to each other
  // That is if a variable x is picked at selection depth i and then assigned,
  // The assignment depth of x is i

  int curr_depth; // Specifies the index of the  current node of Tree_path. Initially curr_depth = -1;
                  

  int empty_cube_index; // specifies the index of a cube of F that is currently empty (it may be an unsatisfiable
  // cube of BCP) 

  BCPfirstNode bcp_first_node; // specifies the type of the first assignment in the queue
                               // DEC_LEFT means that it is a left decision assignment (a new node needs to be created)
                               // DEC_RIGHT means that it is a right decision assignment so no need to create a new node
                               // IMPLIED means that it is an assignment implied by a derived cube, so
                               //      no need to create a new node
                               // SPEC_IMPLIED means that a conflict cube C is derived in the right branch and we
                               //       backtrack to run BCP initialized with unit cube C 

  


  //
  //    Data structures related to signal cubes
  //

  
  
  CUBE Init_sgnl_cubes; // List of initial signal X-cubes 
  
 

  

  DNF New_sgnl_cubes; // New_sgnl_cubes[depth] specify the cubes of F that lose their signal status when 
                     // the node 'Tree_path[depth]' is   deleted. 

  CUBE Sgnl_xvars; // specifies the superset of X-variables present in active signal cubes (a variable of 'Sgnl_xvars' 
                     // maybe currently assigned there may not be currently an active signal cube containing x)


  CUBE Sgnl_yvars; // specifies the superset of Y-variables that are present in active signal cubes. 


  //
  // members specifying the instance

  bool init_empty_cube; // is set to 'true' if the original formula has an empty cube
  bool empty_init_sgnl_cubes; // is set to 'true' when the set of initial signal cubes is empty after running initial BCP
  int num_vars; // number of all variables in F

  int init_ncubes;  // initial number of cubes
  int num_assgn_xvars; // gives the number of X-variables that are currently assigned
  int num_assgn_yvars; // gives the number of Y-variables that are currently assigned
  int num_missing_vars; // gives the number of variables whose literals are not in the formula
  int num_sgnl_yvars; // gives the number of  Y-variables that are present in active signal cubes
  int num_sgnl_xvars; // gives the number of X-variables that are present in active signal cubes
  CUBE Xvars; // specifies quantified variables
  CUBE  Yvars; // specifies non-quantified variables
 

  CUBE Act_sgnl_cubes_cnt0; // 'Act_sgnl_cubes_cnt[var_ind]' gives the number of active SGNL-cubes that contain
                                   // the negative literal of variable 'var_ind+1' (this variable may be in X or Y)
  CUBE Act_sgnl_cubes_cnt1; // 'Act_sgnl_cubes_cnt[var_ind]' gives the number of active SGNL-cubes that contain
                                   // the positive literal of variable 'var_ind+1' 
                               

 

  CUBE Assgn_queue; // used during BCP
  CUBE Unit_xcubes; // contains the indexes of cubes that are currently unit and contain X-variables
  CUBE Unit_ycubes; // contains the indexes of cubes that are currently unit and contain only Y-variables
  CUBE Last_cnfl_cubes; // contains indexes of most recent conflict cubes that may still be unit
  //
  //  Stack related data structures
  //

  // Members  indexed by variables

  DNF Flits0; // Flits0[i] lists the cubes of F with the negative literal of variable i+i
  DNF Flits1; // Flits1[i] lists the cubes of F with the positive  literal of variable i+i
  CUBE Lit_activity0; // Let_activity0[i] - the activity of the neg. lit. of var. 'i+1' in the left part of  D-sequents
  CUBE Lit_activity1;// Let_activity1[i] - activity of the pos. lit. of var. 'i+1'in the left part of D-sequents
  vector <VarState> Var_state; // Var_state[i] contains information about the state of i-th variable
  

  DNF  New_sat_cubes; // New_sat_cubes[i] - contains indexes of the cubes that became satisfied
                      // due to the assignment of depth i
  CUBE All_new_sat_cubes; // contains the set of all cubes that became satisfied in the last BCP run
  CUBE All_new_sgnl_cubes; // contains the set of all cubes that became signal cubes in the last BCP run

  


  CUBE Var_sel_depth ;  // If variable i + 1 is assigned, Var_sel_depth [i] specifies the 
                        //depth at which variable i+1 has been selected for branching. Otherwise,
                        // Var_sel_depth [i] is undefined

 
  CUBE  Red_cube_depth; // If cube F[i] is redundant (respectively satisfied), Red_cube_depth[i] shows the depth up to which
  // F[i] remains redundant (respectively satisfied).
  //Otherwise, the value of Red_cube_depth[i]is undefined

  CUBE Formula_size; // If variable 'i+1' has the right assignment, then Formula_size[i] specifies the number of cubes
                     // in F at the time the left assignment to 'i+1' was made. This information is used when generating
                     // D-sequents for conflict cubes 
 

  // of type vector <type> indexed by cubes

  vector <CubeInfo> Cube_info;


  CCUBE Clean; // 1 (0)  means that an Y-cube F[i] is clean (respectively dirty)

  // basic statistics 
  int num_act_sgnl_cubes; // number of signal cubes that are currently active 
  int curr_num_sgnl_cubes; // number of cubes that currently have the 'signal' status
  int num_confls;   // number of conflicts encounted by CCR (conflicts of SAT-solver are ignored)
  int num_xvar_decs; // number of decisions made on X-variables
  int num_yvar_decs; // number of decisions made on Y-variables
  int num_ress; // number of resolvents generated by CCR
  int init_ycubes; // number of initial ycubes
  int num_ycubes; // number of conflict Y-cubes
  int dirt_ycubes; // init_ycubes + new ycubes that are dirty (i.e. are not noise)
  int max_num_ycubes; // if  'max...  > 0 and  more than 'max...' Y-cubes are generated, the algorithm terminates
  long int num_dseqs; // number of D-sequents generated by CCR
  long int num_frw_blocked_cube_dseqs; // number of D-sequents for blocked cubes computed on the forward motion (i.e. during decision making and BCP)
  long int num_bkw_blocked_cube_dseqs; // number of D-sequents for blocked cubes computed on the backward motion 
  long int num_frw_blocked_cube_dseqs0; // the number of D-sequents for blocked cubes found at the root node
  long int num_left_s_seqs; // the number of s-sequents specifying redundancy of cube when it is satisfied
  long int num_sc_dseqs; // number of sequents due to a cube getting satisfied
  long int num_fc_dseqs; // number of sequents due to a cube getting redundant due to appearance of an empty cube
  long int merging_nodes; // number of nodes where results of left and right branches are merged
  long int hybrid_nodes; // number of hybrid  nodes (when exactly one branch has an empty cube assigned) 
  long int num_resL_dseqs; // number of D-sequents obtained by resolving a left clause-D-sequent with regular D-sequent
  long int num_resR_dseqs; // number of D-sequents obtained by resolving a right clause-D-sequent with regular D-sequent
  long int num_resS_dseqs;// number of D-sequents obtained by resolving two regular D-sequents
  long int num_res_xvars; // number of D-sequents resolved  on X-variables
  long int num_res_yvars; // number of D-sequents resolved on Y-variables
  long int num_res_xvars1; // number of D-sequents produced by pseudo-resolution  on X-variables
  long int num_res_yvars1; // number of D-sequents produced by pseudo-resolution on Y-variables
  int max_num_sgnl_cubes; // maximal number of cubes having signal status over the run of the algorithm
 
  // extra statistis (used when 'extra_stat_flag=true')
  //
  CUBE Num_of_dseqs; // Num_of_dseqs[i] specifies how many D-sequents have been derived for cube F[i]
  int wtch_cube_ind; // if 'wtch_cube_ind >= 0' then 'Dseqs_wtch_cube' 
  DNF Dseqs_wtch_cube; // collects all D-sequents derived for the watched cube
  CUBE Sgnl_stat_depths; // collects all depths up to which the watched cube remains redundant
  CUBE Prvd_red_depths; // depths at which the watched cube was proved redundant
  CUBE Gcnts; // values of gcount at the time D-sequent of the watched cube was derived
  vector <DseqType> Dseq_types; // specifies the types of D-sequents of the watched cube
 

  // Solver's state variables
  bool verbose;
  bool rigor_check; // if true, the program aborts if there is an initial signal Y-cube
  bool just_check_sgnl_cubes; // if true, the program exits after checking for an initial signal Y-cube
  bool prot_media_cubes; // if 'false' only signal cubes are protected
  int selector; // can be used to choose between alternatives 
  bool ign_wrong_yvars; // if 'true', pqe just discards 'wrong' Y-variable (the default is false)
 
  XvarDecMaking xvar_dec_mode; // specifies the decision making mode for X-variables
  YvarDecMaking yvar_dec_mode; // specifies the decision making mode for Y-variables
  long seed_value; // used to initialize random number generator
  long long  max_num_dseqs;
  bool extra_stat_flag; // if set to true, some extra statistics is collected
  bool res_dseqs_flag; // if true prints out some extra info about resolving sequents
  int time_limit; // time limit in seconds
  bool prepr_mode; // exit after preprocessing
  bool prepr_bcp; // perform BCP during preprocessing
  bool prepr_red_cubes; // remove redundant cubes during preprocessing
  bool noise; // if false, only dirty Y-cubes are used in the resulting formula
  long start_gcount; // if start_gcount > 0, the program switches 'verbose' on after 'gcount' exceeds 'start_gcount'
  int range_gcount; // if range_gcount > 0, the program switches 'verbose' in the range from 
                    // 'start_gcount-range_gcount' to 'start_gcount-range_gcount'

  // auxiliary variables
  // CUBE Core_cubes;
  char res_root_name[MAX_NAME];
 
  
 
#include "m1ethods.hh"
#include "m2ethods.hh"
};


//
//  EXTERNAL FUNCTIONS AND GLOBAL VARIABLES
//

extern hsh_tbl hash_table_lits;
extern hsh_tbl hash_table_aux;
extern hsh_tbl hash_table_cubes;
extern hsh_tbl hash_table_cubes1;


void initialize_cube_vects(DNF &A,DNF & Lits0,DNF &Lits1);
int read_dimacs1(FILE *fp,DNF &D,int &num_vars);
void fprint_cube(FILE *fp,CUBE &C);
void print_dnf(DNF &D);
int find_max_var(DNF &D);
void print_dnf(DNF &D,char *fname,int num_vars);
ostream &operator<<(ostream &os,CUBE const &v);
ostream &operator<<(ostream &os,DCUBE const &v);


void add_assumps(MvecLits &Assmps,CUBE &A);
void add_negated_assumps(MvecLits &Assmps,CLAUSE &C);
void accept_new_clauses(Msolver *Slvr,CNF &H);
Msolver *init_sat_solver(int nvars);
void accept_new_clause(Msolver *Slvr,CUBE &C);

void remove_lit(CUBE &E, int lit);
void add_new_lits(CUBE &E,CUBE &C,int lit);
void remove_marked_elems(vector <int> &A);
void get_runtime (double &usrtime, double &systime);
void mark_sat_cubes(CUBE &Cubes);
bool quick_res(CUBE &S1,CUBE &S0,int lit);
void print_mapped_cubes(CubeMap &Cubes);
void print_spaces(int num);
void my_printf(char *format,...);
void my_fprintf(FILE *fp,char *format,...);
void add_array(CUBE &A, CUBE &B);
void print_deps_type(int type);
void read_numbers(const char *fname,CUBE &Sol);
int read_int_num(char *Num_str);

extern long gcount; // counts the algorithm's states
extern bool dec_making_flag;
extern int cnfl_cube_flag;
extern int glob_assgn[];

const int ACTIVE = 0;
const int SATISFIED  = 1;
const int REDUNDANT = 2;
const int CNFL_RED = 3; // used for debugging
const int  MAX_MARKER= 1000000; // used in hash tables
const long long MAX_NUM_DSEQS = 1000000000000LL; // a trillion
const int DECAY_RATE = 1000; // the activity of literals is divided by 2 after derivation of next 'DECAY_RATE' D-sequents
const int ERROR = 100;
const int BUF_SIZE = 10000;
