#include <iostream>
#include <list>
#include <vector>
#include <queue>
#include <set>
#include <map>
#include <stdio.h>
#include <stdlib.h>
#include <strings.h>
#include <algorithm>
#include <ctime>
#include <climits>
#include <cassert>
#include <string.h>

#include "Solver.h"
#include "SimpSolver.h"
#include "System.h"
using namespace std;
#include "e0g-pqe+.hh"
#include "einline.hh"

/*=======================================

  G E N _ C N F L _ C U B E

  This function is meant for
  generating a "conflict cube" when
  the BCP procedure results in a conflict
  ========================================*/
void ExQuant::gen_cnfl_cube(CUBE &E,int &confl_jump_depth,int &sgnl_cube_depth,bool &clean,bool &glb_sgnl)
{

  clean = true;
  glb_sgnl = false;
  
  sgnl_cube_depth = num_vars+1;
  assert(empty_cube_index >= 0);
  assert(curr_depth >= 0);
  assert(Cube_info[empty_cube_index].state != REDUNDANT);
  confl_jump_depth = fnd_cnfl_jmp_dpth(curr_depth);

  E = F[empty_cube_index];
  int empty_cube_depth = fnd_ecube_dpth(E);
  CubeInfo &El = Cube_info[empty_cube_index];
  if (El.signal)
    if (El.depth < sgnl_cube_depth)
      sgnl_cube_depth = El.depth;

  if (El.glb_sgnl)
    glb_sgnl = true;

  clean = clean && Clean[empty_cube_index];

  // mark variables of  E
  hash_table_lits.change_marker();
  hash_table_lits.started_using();

  CUBE &table = hash_table_lits.table;
  int &marker = hash_table_lits.marker;

  for (uint i=0; i < E.size(); i++)
    table[abs(E[i])-1] = marker;

  int depth = curr_depth;
  if (cnfl_cube_flag == 2)   {
    printf("\n-------------\n");
    printf("initial unsat. cube number %d: ",empty_cube_index); 
    cout << E << endl;
  }
  while (true)    {
    Node &n = Tree_path[depth];
    if (empty_cube_depth < depth)	{
      depth--;
      continue;
    }

    if (depth == confl_jump_depth) break;
    num_ress++;
    // produce a new empty cube
    int left_lit = (n.lval == 0)?-n.var:n.var;
    if (n.left_ecube < 0)	{
      p();   
      printf("n.left_ecube = %d\n",n.left_ecube);
      printf("depth = %d\n",depth);
      printf("confl_jump_depth = %d\n",confl_jump_depth);
      printf("empty_cube_depth %d\n",empty_cube_depth);
      exit(ERROR);
    }
    CUBE &C = F[n.left_ecube];   
    assert(Cube_info[n.left_ecube].state != REDUNDANT);
    
    CubeInfo &El1 = Cube_info[n.left_ecube]; 
    if (El1.signal)
      if (El1.depth < sgnl_cube_depth)
        sgnl_cube_depth = El1.depth;

    if (El1.glb_sgnl)
      glb_sgnl = true;
     
    clean = clean && Clean[n.left_ecube];
    if (cnfl_cube_flag == 2)	{
      printf("resolving with cube C%d: ",n.left_ecube); cout << C << endl;
      print_cube_state(n.left_ecube);
      print_Assgns_cut(C);
    }
	
    assert(contains_lit(C,left_lit));
    remove_lit(E,-left_lit);
    add_new_lits(E,C,left_lit);
    
    if (cnfl_cube_flag == 2)
      cout << "the resolvent E is -> " << E << endl;
	
    empty_cube_depth = fnd_ecube_dpth(E);   
    if (empty_cube_depth < confl_jump_depth)
      {// recomputing confl_jump_depth, this needs to be done when encoutering a former decision node n
        // with left branch of type UNSAT (i.e. a conflict cube has been derived in the left branch of 'n').
        confl_jump_depth = fnd_cnfl_jmp_dpth(empty_cube_depth);  
      }  
    depth--;
  } /* while */

  hash_table_lits.done_using();

} /* end of function gen_cnfl_cube */



/*===================================

  A D D _ N E W _ L I T S

  This function addes to E
  the literals of C that are
  different from lit and not
  marked in the hash table

  ASSUMPTIONS:
  1) Cube C contains literal 'lit'

  2) E and C do not have orthogonal
  literals
  ===================================*/
void add_new_lits(CUBE &E,CUBE &C,int lit)
{

  CUBE &table = hash_table_lits.table;
  int &marker = hash_table_lits.marker;
  bool found = false;

  for (uint i=0; i < C.size(); i++) {
    if (C[i] == lit) {
      found = true;
      continue;
    }
    int var_ind = abs(C[i])-1;
    if (table[var_ind] == marker) continue; // Variable 'var_ind+1' been seen before
    table[var_ind] = marker;
    E.push_back(C[i]);
  }

  assert(found);
} /* end of function add_new_lits */


/*===========================

  R E M O V E _ L I T

  This function removes
  from cube E literal 'lit'.

  ASSUMPTIONS:

  Cube E contains exactly one
  literal 'lit'
  ============================*/
void remove_lit(CUBE &E, int lit)
{

  int size = E.size();

  bool found = false;
  for (int i=0; i < size; i++) {
    if (E[i] == lit) {
      found = true;
      continue;
    }

    if (found)
      E[i-1] = E[i];
  }
  if (!found) {
    p();
    printf("explanation: cube E: ");
    cout << E;
    printf(" does not contain literal %d\n",lit);
    exit(ERROR);
  }
  E.resize(size-1); // reduce the size of E by 1

} /* end of function remove_lit */

/*====================================

  F N D _ E C U B E _ D P T H

  This function returns the depth 
  at which a literal of E is set to 1
  that is the closest to 'curr_depth'.

  If E does not have any literals,
  the function returns -1


  ASSUMPTIONS:
  1) all literals of E are assigned
  at depth > curr_depth

  =====================================*/
int ExQuant::fnd_ecube_dpth(CUBE &E)
{
  int max_depth = -1;

  for (uint i=0; i < E.size(); i++) {
    int var_ind = abs(E[i])-1;
    if (Var_state[var_ind].val == 2) {
      p();
      printf("variable %d is not assigned\n",var_ind+1);
      print_alg_state(false);
      printf("curr_depth = %d\n",curr_depth);
      cout << "Empty cube-> " << E << endl;
      print_Assgns_cut(E);
      exit(ERROR);
    }
    if (Var_sel_depth[var_ind] > max_depth)
      max_depth = Var_sel_depth[var_ind];
  }

  return(max_depth);

} /*end of function fnd_ecube_dpth */

/*===========================================

  F N D _ C N F L _ J M P _ D P T H

  This function returns the depth of the
  decision node (left or right) that is
  the closest to 'Tree_path(curr_depth)' 

  If 'Tree_path' does not have decision
  nodes, the function returns -1

  ============================================*/
int ExQuant::fnd_cnfl_jmp_dpth(int starting_depth)
{

  for (int depth = starting_depth; depth >= 0; depth--)
    if (Tree_path[depth].left_ecube < 0) 
      return(depth);
  return(-1);

} /*end of function fnd_cnfl_jmp_dpth */


