#include <iostream>
#include <list>
#include <vector>
#include <queue>
#include <set>
#include <map>
#include <stdio.h>
#include <stdlib.h>
#include <strings.h>
#include <algorithm>
#include <ctime>
#include <climits>
#include <cassert>
#include <string.h>

#include "Solver.h"
#include "SimpSolver.h"
#include "System.h"
using namespace std;
#include "e0g-pqe+.hh" 
#include "einline.hh"



/*==========================================

  S E Q _ G E N _ R E D _ C U B E 1


  This function returns in S a D-sequent of
  cube F[cube_ind1].

  ASSUMPTIONS:
  1) Cube 'F[cube_ind1]' contains literal 'lit1'
  2) Every active cube with literal '-lit1'
  is at distance 2 or more from 'F[cube_ind1]'
  3) S.size() = 0

  ==========================================*/
void ExQuant::seq_gen_red_cube1(CUBE &S,int lit1)
{
  
  hash_table_lits.change_marker();
  hash_table_lits.started_using();

  CUBE *pCubes;

  if (lit1 < 0) pCubes = &Flits1[-lit1-1];
  else pCubes = &Flits0[lit1-1];


  for (uint i=0; i < pCubes->size(); i++)   {
    int cube_ind2 = (*pCubes)[i];
    if (Cube_info[cube_ind2].state == ACTIVE)
      continue; // should be at distance >= 2
    if (Cube_info[cube_ind2].state == SATISFIED) {
      int lit2 = Cause_lit[cube_ind2];
      int var_ind2 = abs(lit2) - 1;
      if (!present_in_table(var_ind2,lit2)) {
        S.push_back(lit2);
        insert_into_table(var_ind2,lit2);  
      }
      continue;
    }

    // F[cube_ind2] is a redundant cube
     
    CUBE &Sa = Act_dseqs[cube_ind2];
    copy_lits(S,Sa);      
  }
  

  hash_table_lits.done_using();
 

} /* end of function seq_gen_red_cube1*/


/*==========================================

  S E Q _ G E N _ R E D _ C U B E 2


  This function returns  a D-sequent of
  cube F[cube_ind] in S.

  In contrast to 'seq_gen_red_cube1' we 
  DO NOT SKIP active cubes with literal '-lit'
  assuming that they have to be at distance >= 2
  from F[cube_ind1]

  ASSUMPTIONS:
  1) Cube 'F[cube_ind1]' contains literal 'lit'
  2) Every cube with literal '-lit' is either
  satisfied or redundant
  3) S.size() = 0

  ==========================================*/
void ExQuant::seq_gen_red_cube2(CUBE &S,int cube_ind1,int lit1)
{

 


  hash_table_lits.change_marker();
  hash_table_lits.started_using();
  
  CUBE *pCubes;

  if (lit1 < 0) pCubes = &Flits1[-lit1-1];
  else pCubes = &Flits0[lit1-1];


  for (uint i=0; i < pCubes->size(); i++)  {
    int cube_ind2 = (*pCubes)[i];
    int dist = distance(F[cube_ind1],F[cube_ind2]);
    if (dist > 1) continue;
    assert(dist == 1);
    if (Cube_info[cube_ind2].state == SATISFIED)  {
      int lit2 = Cause_lit[cube_ind2];
      int var_ind2 = abs(lit2) - 1;
      if (!present_in_table(var_ind2,lit2)) {
        S.push_back(lit2);
        insert_into_table(var_ind2,lit2);  
      }
      continue;
    }


    // F[cube_ind2] - redundant cube
     
    if (Cube_info[cube_ind2].state != REDUNDANT) {
      p();
      print_cube_state(cube_ind2);
      exit(ERROR);
    }
    CUBE &Sa = Act_dseqs[cube_ind2];      
    copy_lits(S,Sa);   
  }

  hash_table_lits.done_using();
 
 
} /* end of function seq_gen_red_cube2 */


/*========================================================

  S E Q _ G E N _ E C U B E

  Node 'n' is the current node.
  This function  generates a sequent  S 
  for the empty cube of one of the branches.

  ASSUMPTIONS:
  1) n is not a conflict cube

  =======================================================*/
void ExQuant::seq_gen_ecube(Node &n)
{
  
  // add literals that mades cubes with variable n.var
  // satisfied or redundant
 

  int count = 0;
  if (n.left_ecube >= 0) count++;
  if (n.right_ecube >= 0) count++;
  assert(count != 2);
  if (count == 0) return;
 
  hybrid_nodes++;
     
  CUBE S;
  int lit = 0;
  int cube_ind = -1;
  bool recompute_flag,replace_flag;
  if (n.left_ecube >= 0)   {
    if (Cube_info[n.left_ecube].signal == 0) return;
    lit = (n.lval == 1)?n.var:-n.var; 
    cube_ind = n.left_ecube; 
    recompute_flag = false; // left child cube is in the state satisfied 
    replace_flag = false; // changing TRUE to FALSE!!!!
  }
  else // n.left_ecube < 0)
    if (n.right_ecube >= 0)     {
      if (Cube_info[n.right_ecube].signal == 0) return;
      lit = (n.lval == 1)?-n.var:n.var; 
      cube_ind = n.right_ecube; 
      recompute_flag = true; // right child cube is in the state active 
      replace_flag = false;
    }
    else assert(false);

  
  
  seq_gen_red_cube2(S,cube_ind,lit);
   
  num_bkw_blocked_cube_dseqs++;

  int L = max_cube_recovery_depth(S);
  set_new_act_dseq(cube_ind,S,replace_flag,UNRES);
  check_red_cube_in(cube_ind,lit,L,recompute_flag);
 
  

} /* end of function seq_gen_ecube */
