#include <iostream>
#include <list>
#include <vector>
#include <queue>
#include <set>
#include <map>
#include <stdio.h>
#include <stdlib.h>
#include <strings.h>
#include <algorithm>
#include <ctime>
#include <climits>
#include <cassert>
#include <string.h>

#include "Solver.h"
#include "SimpSolver.h"
#include "System.h"
using namespace std;
#include "e0g-pqe+.hh"
#include "einline.hh"
	


/*============================================

  N U M _ S G N L _ C U B E S

  This functions returns the number of
  signal cubes among 'Cubes' 

  ASSUMPTIONS: 
  1) Elements of Cube_info[cube_ind].signal
  are assigned proper values

  =============================================*/
int ExQuant::num_sgnl_cubes(CUBE &Cubes)
{
  int count = 0;
  for (uint i=0; i < Cubes.size(); i++) {
    int cube_ind = Cubes[i];
    CubeInfo &El = Cube_info[cube_ind];
    if (El.state == REDUNDANT) continue;
    if (El.signal) count++;
  }

  return(count);
} /* end of function num_sgnl_cubes */

/*===========================================

  S E T _ I N I T _ C O U N T S

  This function sets the initial
  counts of 'act_X_cubes_cnt'

  ASSUMPTIONS:

  currently 'act_X_cubes_cnt[var_ind] = -1'


  ===========================================*/
void ExQuant::set_init_counts()
{

  for (int i=0; i < num_vars; i++)    {
    Act_sgnl_cubes_cnt0[i] = num_sgnl_cubes(Flits0[i]);
    Act_sgnl_cubes_cnt1[i] = num_sgnl_cubes(Flits1[i]);

    if (Act_sgnl_cubes_cnt0[i] + Act_sgnl_cubes_cnt1[i] > 0) {
      if (Var_state[i].xvar)
        add_xvar_to_array(i);
      else
        add_yvar_to_array(i);
    }
  }


} /* end of function set_init_counts */



/*=====================================

  F O R M _ X V A R S _ Y V A R S


  ASSUMPTIONS:
 
  1) Yvars is already formed but may
  contain missing variables

  2) Xvars is empty

  =====================================*/
void ExQuant::form_Xvars_Yvars()
{

  // remove missig variables from Yvars

  int shift = 0;

  for (uint i=0; i < Yvars.size(); i++) {
    int var_ind = Yvars[i]-1;
    if (var_ind >= num_vars) {
      p();
      exit(ERROR);
    }
    if (missing_var(var_ind)) {
      shift++;
      continue;}

    if (shift > 0)
      Yvars[i-shift] = Yvars[i];

  }

  if (shift > 0) Yvars.resize(Yvars.size() - shift);


  // mark Yvars 

  for (uint i=0; i < Yvars.size(); i++) {
    int var_ind = Yvars[i]-1;
    Var_state[var_ind].xvar = 0;}


  // form Xvars
  Xvars.clear();

  for (int i=0; i < num_vars; i++) {
    if (missing_var(i)) {
      num_missing_vars++;
      continue;
    }
    if (Var_state[i].xvar) Xvars.push_back(i+1);
  }


} /* end of function form_Xvars_Yvars */

/*======================================

  I N I T _ S O L V E R _ S T A T E

  =======================================*/
void ExQuant::init_solver_state()
{
  verbose = false;
  selector = 0;
  xvar_dec_mode = XVAR_MAX_ACT_SGNL_CUBES;
  yvar_dec_mode = YVAR_LIT_ACTIVITY5;
  seed_value = 0;
  max_num_dseqs = MAX_NUM_DSEQS;
  dec_making_flag = false;
  cnfl_cube_flag = 0;  
  strcpy(res_root_name,"res");
  extra_stat_flag = false;
  wtch_cube_ind = -1;
  empty_init_sgnl_cubes = false;
  prot_media_cubes = true;
  rigor_check = true;
  just_check_sgnl_cubes = false;
  ign_wrong_yvars = false;
  res_dseqs_flag = false;
  max_num_ycubes = -1;
  first_yvar_assigned = 0;
  time_limit = 1000000;
  prepr_mode = false;
  prepr_bcp = false;
  prepr_red_cubes = false;
  noise = true;
  start_gcount = -1;
  range_gcount = -1;
  max_gcount = -1;
} /* end of function init_solver_state */
