#include <iostream>
#include <list>
#include <vector>
#include <set>
#include <map>
#include <string>
#include <queue>
#include <algorithm>
//
#include <stdio.h>
#include <assert.h>
#include <string.h>


#include "Solver.h"
#include "SimpSolver.h"
#include "System.h"
using namespace std;

#include "e0g-pqe+.hh"
#include "einline.hh"


//
//  Global variables
//

hsh_tbl hash_table_lits;
hsh_tbl hash_table_aux;
hsh_tbl hash_table_cubes;
hsh_tbl hash_table_cubes1;

/*===================

  M A I N

  ==================*/
int main(int argc,char *argv[])
{


  ExQuant ExQ;

  ExQ.time_limit = -1;
  ExQ.max_new_ycubes = -1;
  
  if (argc == 1)  {
    ExQ.print_header();
    exit(0);
  }

  if (argc > 4)   {
    printf("wrong number of arguments: %d\n",argc);
    exit(ERROR);
  }

  ExQ.read_input(argv[1]);
  ExQ.init_solver_state();
  ExQ.read_parameters(argc,argv);

  if (ExQ.time_limit > 0) 
    Minisat::limitTime(ExQ.time_limit+10);
  
  double usrtime0=0.,systime0=0.;
  double usrtime=0.,systime=0.;
  get_runtime (usrtime0, systime0);

  ExQ.print_version();
  bool fin1,fin2;
  fin1 = ExQ.init();
  if (!fin1) fin2 = ExQ.eg_pqe_plus();
  ExQ.print_stat1();
  get_runtime (usrtime, systime);
  printf("total time is %.2f sec.\n",usrtime-usrtime0);
  char Fname[MAX_NAME];
  strcpy(Fname,argv[1]);
  strcat(Fname,".res.cnf");
  print_dnf(ExQ.Sol,Fname,ExQ.num_vars);
  if (fin1) {
    printf("solved by preprocessing\n");
    exit(0);
  }
  if (fin2) {
    printf("successful termination\n");   
    exit(0);
  }
  else {
    printf("not finished\n");
    exit(1);
  }

}/* end of function main */


//
//  EG_PQE_PLUS
//
bool ExQuant::eg_pqe_plus()
{

  for (uint i=0; i < Sol.size(); i++)
    if (Sol[i].size() == 0) return(true);

  bool ok;
  while (Prim_sgnl_cubes.size() > 0) {
    int cube_ind = Prim_sgnl_cubes.back();
    printf("================\n");
    printf("proving redundancy of F[%d]\n",cube_ind);
    print_cube1(F[cube_ind]);
    printf("gcount = %ld\n",gcount);
    printf("================\n");
    CubeInfo &El = Cube_info[cube_ind];
    if (El.state != REDUNDANT) 
      ok = prove_red(cube_ind);
    else  {
      if (Red_cube_depth[cube_ind] >= 0) {
        p();
        print_cube_state(cube_ind);
        exit(ERROR);
      }
      ok = true;
    }
    
    upd_prim_cubes_info(cube_ind);    
    if (!ok) return(false);
  }

  return(true);

} /* eg_pqe_plus */


//
//  READ_INPUT
//
void ExQuant::read_input(char *root)
{  
  char fname[MAX_NAME];
  strcpy(fname,root);

  // read CNF formula
  strcat(fname,".cnf");
  FILE *fp1 = fopen(fname,"r");
  if (fp1 == NULL) {
    p();
    exit(ERROR);
  }
  read_dimacs1(fp1,F,num_vars);
  
  // read list of Y-variables
  fname[0] = 0;
  strcpy(fname,root);
  strcat(fname,".txt");
  read_numbers(fname,Yvars); 
  
  
  // read the list of cubes to be taken out of the formula
  fname[0] = 0;
  strcpy(fname,root);
  strcat(fname,".clauses");
  read_numbers(fname,Trgs);
  
}  /* read_input */


//
//   INIT
//
bool ExQuant::init()
{

  
  Xvar.assign(num_vars,1);
  for (uint i=0; i < Yvars.size(); i++)
    Xvar[Yvars[i]-1] = 0;

  SCUBE S;
  for (uint i=0; i < Trgs.size(); i++)
    S.insert(Trgs[i]);

  for (uint i=0; i < F.size(); i++)
    if (S.find(i) == S.end()) F2.push_back(F[i]);


  for (int i=0; i < num_vars; i++)    {
    CUBE  Dummy;
    Flits0.push_back(Dummy);
    Flits1.push_back(Dummy);
  }

  // initialize literal counts  
  initialize_cube_vects(F,Flits0,Flits1);

  Slv = init_sat_solver(num_vars);
  accept_new_clauses(Slv,F);

  new_ycubes = 0;
  rem_bnd_pnts = 0;
  unrem_bnd_pnts = 0;
  
  // initialize DS-PQE
 

  for (uint i=0; i < Trgs.size(); i++) {
    Init_sgnl_cubes.push_back(Trgs[i]);
    Prim_sgnl_cubes.push_back(Trgs[i]);
  }

  initialize();
  
  bool solved = assgn_init_prot_cubes();
  return(solved);
  
} /* init */

//
//  PRINT_STAT1
//
void ExQuant::print_stat1()
{
  printf("#removable boundary points %d\n",rem_bnd_pnts);
  printf("#unremovable boundary points %d\n",unrem_bnd_pnts);
  print_solver_state();
  print_stat();         

} /* print_stat1 */
