#include <iostream>
#include <list>
#include <vector>
#include <set>
#include <map>
#include <string>
#include <queue>
#include <algorithm>
//
#include <stdio.h>
#include <assert.h>
#include <string.h>


#include "Solver.h"
#include "SimpSolver.h"
#include "System.h"
using namespace std;

#include "e0g-pqe+.hh"
#include "einline.hh"

//
//  PRINT_VERSION
//

void ExQuant::print_version()
{
  printf("-------------------------\n");
  printf("  EG-PQE+, version 1.0\n");
  printf("-------------------------\n");
} /* print_version */

//
//   PRINT_HEADER
//
void ExQuant::print_header()
{
  print_version();
  printf("Synopsis: \n");
  printf("  eg_pqe+ root ['d'mlrst] ['e'num] ['g' gcnt rng] ['G'gcnt] ['m'gcnt] ['t'val] [v]\n");
  printf("  'd'          - set the decision making mode for quant. vars \n");
  printf("  'e'num   - set number of ycubes to generate\n");
  printf("  'G'gcnt      -   switch on verbose after 'gcount' reaches value 'gcnt'\n");
  printf("  'g' gcnt rng - switch on verbose when 'gcount == gcnt' and\n");
  printf("                 switch off when 'gcount == gcnt+rng'\n");
  printf("  'm'gcnt   - terminate after gcount reach gcnt\n");
  printf("  't'val   - sets the time limit (in seconds)\n");
  printf("  'v'      - verbose output\n");
} /* print_header */

//
//   READ_PARAMETERS
//
void ExQuant::read_parameters(int argc,char *argv[])
{ 
  for (int i=2; i < argc; i++) {
    switch (argv[i][0]) {
    case 'd':
      if (argv[i][1] == 0) {
        p();
        printf("no choice of dec. mak. heur. for int. vars is given\n");
        exit(ERROR);
      }
      switch(argv[i][1]) {
      case 'm':
        xvar_dec_mode = XVAR_MAX_ACT_CUBES;
        break;
      case 'l':
        xvar_dec_mode = XVAR_LIT_ACTIVITY;
        break;
      case 'r':
        xvar_dec_mode = XVAR_RAND;
        break;
      case 's':
        xvar_dec_mode = XVAR_TRIV;
        break;
      case 't':
        xvar_dec_mode = XVAR_MAX_ACT_SGNL_CUBES;
        break;
      default:
        p();
        printf("wrong dec. mak. heur. for X-vars\n");
        exit(ERROR);
      }
      break;
    case 'e':
      max_new_ycubes = atoi(argv[i]+1);
      break;
    case 'G':
      start_gcount = read_int_num(argv[i]+1);
      assert(start_gcount >=0);
      break;
    case 'g':
      assert(i + 2 < argc);
      start_gcount = read_int_num(argv[i+1]);
      assert(start_gcount >= 0);
      range_gcount = read_int_num(argv[i+2]);
      assert(range_gcount > 0);
      i += 2;
      break;
    case 'm':
      max_gcount = atoi(argv[i]+1);
      break;
    case 't':
      time_limit = atoi(argv[i]+1);
      if (time_limit <= 0)   {
        p();
        printf("wrong time limit %s\n",argv[i]);
        exit(ERROR);
      }
      break;
    case 'v':
      verbose = true;
      break;    
    default:
      p();
      printf("wrong parameter %s\n",argv[i]);
      exit(ERROR);
    }
  }
} /* read_parameters */
