#include <iostream>
#include <list>
#include <vector>
#include <queue>
#include <set>
#include <map>
#include <stdio.h>
#include <stdlib.h>
#include <strings.h>
#include <string.h>
#include <time.h>
#include <algorithm>
#include <ctime>
#include <climits>
#include <cassert>
#include <sys/time.h>
#include <sys/resource.h>
#include <unistd.h>

#include "Solver.h"
#include "SimpSolver.h"
#include "System.h"
using namespace std;
#include "e0g-pqe+.hh"
#include "einline.hh"

/*================================================

  P R I N T _ X V A R _ D E C _ M O D E

  ================================================*/
void ExQuant::print_xvar_dec_mode()
{
  switch(xvar_dec_mode)    {
  case XVAR_RAND:
    printf("XVAR_RAND\n");
    break;
  case XVAR_TRIV:
    printf("XVAR_TRIV\n");
    break;
  case XVAR_MAX_ACT_CUBES:
    printf("XVAR_MAX_ACT_CUBES\n");
    break;
  case XVAR_MAX_ACT_SGNL_CUBES:
    printf("XVAR_MAX_ACT_SGNL_CUBES\n");
    break;
  case XVAR_LIT_ACTIVITY:
    printf("XVAR_LIT_ACTIVITY\n");
    break;
  default:
    p();
    printf("wrong mode!!\n");
    exit(ERROR);
  }
} /* end of function print_xvar_dec_mode */

/*=================================================

  P R I N T _ Y V A R  _ D E C _ M O D E

  ==================================================*/
void ExQuant::print_yvar_dec_mode()

{  
  printf("decision making mode of Y-variables: ");
  switch (yvar_dec_mode) {
  case YVAR_RAND:
    printf("YVAR_RAND\n");
    break;
  case YVAR_TRIV:
    printf("YVAR_TRIV\n");
    break;
  case YVAR_MAX_ACT_CUBES:
    printf("YVAR_MAX_ACT_CUBES\n");
    break;
  case YVAR_LIT_ACTIVITY3: 
    printf("YVAR_LIT_ACTIVITY3\n");
    break;
  case YVAR_LIT_ACTIVITY5: 
    printf("YVAR_LIT_ACTIVITY5\n");
    break;
  default:
    p();
    printf("wrong mode!!\n");
    exit(ERROR);
  }
} /* end of function print_yvar_dec_mode */
/*=======================================

  P R I N T _ I N F O

  ========================================*/
void ExQuant::print_info(int final_size)
{

  // initial instance information
  printf("-------------\n");
  my_printf((char *) "Initial formula Exist(X)[F(X,Y)]: %m vars, %m clauses\n",num_vars, init_ncubes);
  int num_xvars = (int) Xvars.size();
  int num_yvars = (int) Yvars.size();
  my_printf((char *) "    (|X| = %m , |Y| =  %m, %m variables are missing)\n",num_xvars,num_yvars,num_vars-num_xvars-num_yvars);
  
  my_printf((char *) "The size of the final formula F*(Y) is  %m clauses \n",final_size);
  my_printf((char *) "The total number of ycubes is %m\n",num_ycubes);
  printf("---------------\n");
  print_solver_state();
  printf("--------------\n");
  print_stat();
  my_printf((char *) "gcount = %m\n",gcount);
  my_printf((char *) "num_res_xvars1 = %m\n",num_res_xvars1);
  my_printf((char *) "num_res_yvars1 = %m\n",num_res_yvars1);
 
} /* end of function print_info */

/*============================

  P R I N T _ S T A T

  ===========================*/
void ExQuant::print_stat()

{
  my_printf((char *) "total number of  conflicts = %m\n",num_confls);
  printf("the share of blocked cubes found at level -1 is %.1f%%\n",num_frw_blocked_cube_dseqs0*100./init_ncubes);
  my_printf((char *) "total number of non-conflict nodes of the search tree is %m\n",merging_nodes);
  my_printf((char *) "total number of hybrid nodes is %m\n",hybrid_nodes);
  my_printf((char*) "total number of  decisions on Y-vars is %m\n",num_yvar_decs);
  my_printf((char*) "total number of  decisions on X-vars is %m\n",num_xvar_decs);
  my_printf((char *) "total number of resolutions of clauses = %m\n",num_ress);
 
  my_printf((char *) "total number of D-sequents of forward (backward) blocked cubes = %m (%m)\n",
	    num_frw_blocked_cube_dseqs,num_bkw_blocked_cube_dseqs);
  my_printf((char *) "total number of D-sequents produced by resolving other D-sequents = %m\n",total_num_res_dseqs());
  my_printf((char *) "total number of left branch S-sequents built explicitly %m\n",num_left_s_seqs);
  my_printf((char *) "total number of D-sequents = %m\n",num_dseqs+num_frw_blocked_cube_dseqs0); 
  my_printf((char *) "the sum of  blocked cube D-seqs, S-seqs, lifed D-seqs and resovlents = %m\n", 
	    num_blocked_cube_dseqs()+total_num_res_dseqs()+num_left_s_seqs);
  my_printf((char *) "total number of S-sequents = %m\n",num_sc_dseqs);
  my_printf((char *) "total number of falsified cube sequents = %m\n",num_fc_dseqs);
  my_printf((char *) "number D-sequents produced by resolving out X_variables = %m\n",num_res_xvars);
  my_printf((char *) "number D-sequents produced by resolving out Y-variables = %m\n",num_res_yvars);
  my_printf((char *) "maximum number of signal cubes = %m\n",max_num_sgnl_cubes);
 
} /* end of function print_stat */




/*=======================================

  P R I N T _ S O L V E R _ S T A T E

  =======================================*/
void ExQuant::print_solver_state()
{

  printf("decision making mode of X-variables: ");
  print_xvar_dec_mode();
  print_yvar_dec_mode();
   
  printf("seed_value = %ld\n",seed_value); 
  printf("Time_limit = %d\n",time_limit);  
}/* end of function print_solver_state */
